;; ##################################################################################
;; # ============================================================================== #
;; # MrEd Designer - callback-code.ss                                               #
;; # http://mreddesigner.lozi.org                                                   #
;; # Copyright (C) Peter Ivanyi, 2007                                               #
;; # ============================================================================== #
;; #                                                                                #
;; # This program is free software; you can redistribute it and/or                  #
;; # modify it under the terms of the GNU General Public License                    #
;; # as published by the Free Software Foundation; either version 2                 #
;; # of the License, or (at your option) any later version.                         #
;; #                                                                                #
;; # This program is distributed in the hope that it will be useful,                #
;; # but WITHOUT ANY WARRANTY; without even the implied warranty of                 #
;; # MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the                  #
;; # GNU General Public License for more details.                                   #
;; #                                                                                #
;; # You should have received a copy of the GNU General Public License              #
;; # along with this program; if not, write to the Free Software                    #
;; # Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.    #
;; #                                                                                #
;; ##################################################################################

(module mreddesigner-callback-code mzscheme
  (require (lib "class.ss")
           (file "graph-editor/graph-editor.ss")
           (file "mreddesigner-misc.ss")
           (file "mreddesigner-widgets.ss")
           (file "mreddesigner-callback.ss")
           (file "mreddesigner-callback-edit.ss")
  )
  
        ; check whether we set-list and set-selection of the
        ; same widget, if yes then first set-list then the selection
        
        ; no loops are allowed
  
  (define (make-action action dst-symbol dst-type value)
    (cond
      ((equal? action "enable")    (list 'send dst-symbol 'enable #t))
      ((equal? action "disable")   (list 'send dst-symbol 'enable #f))
      ((equal? action "show")      (list 'send dst-symbol 'show #t))
      ((equal? action "hide")      (list 'send dst-symbol 'show #f))
      ((equal? action "check")     (list 'send dst-symbol 'set-value #t))
      ((equal? action "uncheck")   (list 'send dst-symbol 'set-value #f))
      ((equal? action "toggle")    
       (list 'send dst-symbol 'set-value (list 'not (list 'send dst-symbol 'get-value)))
      )
      ((equal? action "set-label") (list 'send dst-symbol 'set-label value))
      ((equal? action "set-text")  (list 'send dst-symbol 'set-value value))
      ((equal? action "set-value") (list 'send dst-symbol 'set-value value))
      ((equal? action "set-selection") (list 'send dst-symbol 'set-selection value))
      ((equal? action "set-list")
       (let
         ((type dst-type))
         (cond
           ((equal? type 'choice)
            (list 'begin
                  (list 'send dst-symbol 'clear)
                  (list 'for-each 
                        (list 'lambda '(val) (list 'send dst-symbol 'append 'val)) 
                        (list 'quote value))
            )
           )
           ((equal? type 'list-box)
            (list 'send dst-symbol 'set (list 'quote value))
           )
           (else
            (error "make-action: unhandled control type: " type)
           )
         )
       )
      )
      (else
        (error "make-action: unhandled action: " action)
      )
    )
  )
  
  ; generates the callback itself
  (define (make-function callback-info src-id src-type dst-id dst-type)
    (let*
      ((src         (list-ref callback-info 0))
       (is?         (list-ref src 0))
       (cond-value  (list-ref src 1))
       (dst         (list-ref callback-info 1))
       (action      (list-ref dst 0))
       (value       (list-ref dst 1))
       (dst-symbol  (string->symbol dst-id))
       (func-action (make-action action dst-symbol dst-type value))
      )
      (case src-type
        ((button)
         func-action
        )
        ((check-box)
         (list 'let
               '((res (send w get-value)))
               (if (string-ci=? "true" cond-value)
                 (list 'if 'res func-action)
                 (list 'if '(not res) func-action)))
        )
        ((radio-box)
         (list 'let
               '((n (send w get-selection)) (str (send w get-item-plain-label n)))
               (list 'if (list 'and '(string? str) 
                               (if is?
                                 (list 'string-ci=? 'str cond-value)
                                 (list 'not (list 'string-ci=? 'str cond-value))))
                     func-action))
        )
        ((choice)
         (list 'let
               '((str (send w get-string-selection)))
               (list 'if (list 'and '(string? str) 
                               (if is?
                                 (list 'string-ci=? 'str cond-value)
                                 (list 'not (list 'string-ci=? 'str cond-value))))
                     func-action))
        )
        ((list-box)
         (list 'let
               '((str (send w get-string-selection)))
               (list 'if (list 'and '(string? str) 
                               (if is?
                                 (list 'string-ci=? 'str cond-value)
                                 (list 'not (list 'string-ci=? 'str cond-value))))
                     func-action))
        )
        ((text-field)
         (list 'let
               '((type (send e get-event-type)))
               (list 'if (list 'and '(equal? type 'text-field-enter)
                               (if is?
                                 (list 'string-ci=? '(send w get-value) cond-value)
                                 (list 'not (list 'string-ci=? '(send w get-value) cond-value))))
                     func-action))
        )
        ((slider)
         (list 'if (if is? 
                     (list 'equal? '(send w get-value) cond-value)
                     (list 'not (list 'equal? '(send w get-value) cond-value)))
               func-action)
        )
        (else
         (error "make-function: Unhandled source widget type: " src-type)
        )
      )
    )
  )
  
  ; generates functions to pass callback control further down
  (define (make-function-pass callback-info dst-id dst-type)
    (let*
      ((src         (list-ref callback-info 0))
       (is?         (list-ref src 0))
       (cond-value  (list-ref src 1))
       (dst         (list-ref callback-info 1))
       (action      (list-ref dst 0))
       (value       (list-ref dst 1))
      )
      (cond
        ((and (equal? dst-type 'radio-box)
              (member action '("set-list" "set-selection")))
         (list (list (callback-name dst-id) 
                     (string->symbol dst-id)
                     '(new control-event% (event-type 'radio-box))))
        )
        ((and (equal? dst-type 'list-box)
              (member action '("set-list" "set-selection")))
         (list (list (callback-name dst-id) 
                     (string->symbol dst-id)
                     '(new control-event% (event-type 'list-box))))
        )
        ((and (equal? dst-type 'choice)
              (member action '("set-list" "set-selection")))
         (list (list (callback-name dst-id) 
                     (string->symbol dst-id)
                     '(new control-event% (event-type 'choice))))
        )
        ((and (equal? dst-type 'check-box)
              (member action '("check" "uncheck" "toggle")))
         (list (list (callback-name dst-id) 
                     (string->symbol dst-id)
                     '(new control-event% (event-type 'check-box))))
        )
        ((and (equal? dst-type 'text-field)
              (equal? action "set-text"))
         (list (list (callback-name dst-id) 
                     (string->symbol dst-id)
                     '(new control-event% (event-type 'text-field-enter))))
        )
        ((and (equal? dst-type 'slider)
              (equal? action "set-value"))
         (list (list (callback-name dst-id) 
                     (string->symbol dst-id)
                     '(new control-event% (event-type 'slider))))
        )
        (else
         '()
        )
      )
    )
  )
  
  
  (provide callback-name)
  (define (callback-name id)
    (string->symbol (string-append id "-callback"))
  )
  
  
  ; generate the callback function and also call all of its descendents
  (provide mreddesigner-callback-generate)
  (define (mreddesigner-callback-generate id)
    (let*
      ; get the node
      ((node (send graph-editor-canvas node-get-by-name id)))
      (if node
        (let*
          ((type      (node-property-get node 'type))
           (tab-out   (send node tab-out-ref 0)) ; the the only out tab
           (lines     (send tab-out get-lines)) ; the lines connected to the out tab
          )
          ; no lines are connected
          (if (null? lines)
            '(lambda (w e) (void))
            (let*
              ; the tabs at the other end of lines
              ((act-tab-in   (map (lambda (line) (send line get-target)) lines))
               ; the action nodes 
               (act-nodes    (map (lambda (tab) (send tab get-node)) act-tab-in))
               ; the actions in the nodes
               (actions     (map (lambda (node) (node-property-get node 'callback-action)) act-nodes))
               ; the out tabs of the actions, if there is action
               (act-tab-out (map (lambda (node action) 
                                   (if action
                                     (send node tab-out-ref 0) #f)) 
                                 act-nodes actions))
               ; the outgoing lines from the out-tab
               (line-out    (map (lambda (tab)
                                   (if tab
                                     (car (send tab get-lines)) #f))
                                 act-tab-out))
               ; the other ends of the lines, tabs
               (dst-tabs    (map (lambda (line)
                                   (if line
                                     (send line get-target) #f)) 
                                 line-out))
               ; nodes belonging to destination tabs
               (dst-nodes   (map (lambda (tab)
                                   (if tab
                                     (send tab get-node) #f)) 
                                 dst-tabs))
               ; the destination node ids
               (dst-ids     (map (lambda (node)
                                   (if node 
                                     (send node get-name) #f))
                                 dst-nodes))
               ; the destination types
               (dst-types   (map (lambda (dst)
                                   (if dst
                                     (node-property-get dst 'type) #f))
                                 dst-nodes))
              )
              ; generate the lambda function
              (append '(lambda (w e))
                      (apply append
                             (map
                               (lambda (dst-id dst-type action)
                                 (if dst-id
                                   (list (make-function action id type dst-id dst-type))
                                   '()
                                 )
                               )
                               dst-ids dst-types actions))
                      ; if the destination can affect others then 
                      ; call the destination callback as well
                      (remove-duplicates
                        (apply append
                               (map
                                 (lambda (dst-id dst-type action)
                                   (if action
                                     (make-function-pass action dst-id dst-type)
                                     '()
                                   )
                                 )
                                 dst-ids dst-types actions)))))))
        '(lambda (w e) (void))
      )
    )
  )
  
  
) ; end of module


