;; ##################################################################################
;; # ============================================================================== #
;; # MrEd Designer - default-values.ss                                              #
;; # http://mreddesigner.lozi.org                                                   #
;; # Copyright (C) Lozi Jean-Pierre, 2004 - mailto:jean-pierre@lozi.org             #
;; # Copyright (C) Peter Ivanyi, 2007                                               #
;; # ============================================================================== #
;; #                                                                                #
;; # This program is free software; you can redistribute it and/or                  #
;; # modify it under the terms of the GNU General Public License                    #
;; # as published by the Free Software Foundation; either version 2                 #
;; # of the License, or (at your option) any later version.                         #
;; #                                                                                #
;; # This program is distributed in the hope that it will be useful,                #
;; # but WITHOUT ANY WARRANTY; without even the implied warranty of                 #
;; # MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the                  #
;; # GNU General Public License for more details.                                   #
;; #                                                                                #
;; # You should have received a copy of the GNU General Public License              #
;; # along with this program; if not, write to the Free Software                    #
;; # Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.    #
;; #                                                                                #
;; ##################################################################################

(module mreddesigner-default-values mzscheme
  (require (lib "class.ss")  
           (lib "mred.ss" "mred")
           (file "mreddesigner-widgets.ss")
  )
  
  ; icons
  (define k-icon-table (make-hash-table 'equal))
  (for-each
    (lambda (elt)
      (hash-table-put! k-icon-table
                       (string->symbol (substring elt 0 (- (string-length elt) 4)))
                       (make-object bitmap% 
                                   (build-path "images" "icons" "16x16" elt)
                                   'png #f))
    )
    (map (lambda (p) (path->string p))
         (directory-list (path->string (build-path "images" "icons" "16x16"))))
  )
  (provide icon-get)
  (define (icon-get type)
    (hash-table-get k-icon-table type #f)
  )
  
  ; changed parameter
  (provide changed! changed?)
  (define no-change-since-last-save #t)
  (define (changed! bool)
    (set! no-change-since-last-save bool)
  )
  (define (changed?)
    no-change-since-last-save
  )
  
  ;;========================================================
  ;; * widgets in general
  ;;========================================================

  ; this variable stores the widget id on which we clicked in the preview
  (define widget-clicked #f)
  
  ; whether the object is a container
  (provide container?)
  (define (container? type)
    (member type '(frame pane panel 
                   horizontal-pane horizontal-panel 
                   vertical-pane vertical-panel tab-panel 
                   group-box-panel grow-box-spacer-pane)))

  (provide widget-default-values)
  (define (widget-default-values type)
    (cond
      ((eq? type 'project)              project-default-values)
      ((eq? type 'frame)                frame-default-values)
      ((eq? type 'button)               button-default-values)
      ((eq? type 'radio-box)            radio-box-default-values)
      ((eq? type 'check-box)            check-box-default-values)
      ((eq? type 'pane)                 pane-default-values)
      ((eq? type 'horizontal-pane)      horizontal-pane-default-values)
      ((eq? type 'vertical-pane)        vertical-pane-default-values)
      ((eq? type 'panel)                panel-default-values)
      ((eq? type 'horizontal-panel)     horizontal-panel-default-values)
      ((eq? type 'vertical-panel)       vertical-panel-default-values)
      ((eq? type 'tab-panel)            tab-panel-default-values)
      ((eq? type 'group-box-panel)      group-box-panel-default-values)
      ((eq? type 'grow-box-spacer-pane) grow-box-spacer-pane-default-values)
      ((eq? type 'message)              message-default-values)
      ((eq? type 'slider)               slider-default-values)
      ((eq? type 'gauge)                gauge-default-values)
      ((eq? type 'list-box)             list-box-default-values)
      ((eq? type 'choice)               choice-default-values)
      ((eq? type 'text-field)           text-field-default-values)
      ((eq? type 'canvas)               canvas-default-values)
      (else
        (error "Unknown type: " type)
      )
    )
  )
  
  ;;========================================================
  ;; * project
  ;;========================================================

  (provide project-default-values)
  (define project-default-values '((type project)
                                   (label "Untitled Project")
                                   (path ""))
  )
  
  ;;========================================================
  ;; * frame
  ;;========================================================
  
  ; This is a special class for previewing frames (windows).
  ; It registers its position and its size in the data structure,
  ; so it is possible to restore.
  (define mred-frame%
    (class frame%
      ; id of the frame in the data list
      (init-field 
        (mred-id #f)
        (preview-callback #f)
      )
      (define/public (mred-id-set! id)
        (set! mred-id id))
      (define/public (mred-id-get)
        mred-id)
    
      (define/augment (on-close)
        (let ((selected (send hierarchy-widget get-selected)))
          (if selected
            (let 
              ((ptr (send hierarchy-widget item-property-get selected 'pointer #f)))
              ; record that the window is hidden
              (send hierarchy-widget item-property-set! mred-id 'hidden #t #f)
              ; if this is the selected widget in the hierarchy window then update the buttons
              (if (equal? ptr this)
                (begin
                  (send hierarchy-show-button enable #t)
                  (send hierarchy-hide-button enable #f))))))
      )
      
      ; registering functions
      (override on-move on-size)
      (define (on-move x y)
        ; store in the hierarchy only if the window is visible
        ; This check is due to a strange error on my Debian Linux
        ; The order of events is not correct somehow, I close (destroy)
        ; the window then I receive this on-move event ...
        (if (send this is-shown?)
          (begin
            (send hierarchy-widget item-property-set! mred-id 'x x #f)
            (send hierarchy-widget item-property-set! mred-id 'y y #f)
          )
        )
        ; notify the property window, but only if this frame is selected 
        (let
          ((selected (send hierarchy-widget get-selected)))
          (if (equal? selected mred-id)
            (begin
              (send properties-panel set-value 'x x)
              (send properties-panel set-value 'y y)
            )
          )
        )
      )
      
      (define (on-size w h)
        ; store in the hierarchy
        (if (send this is-shown?)
          (begin
            (send hierarchy-widget item-property-set! mred-id 'width w #f)
            (send hierarchy-widget item-property-set! mred-id 'height h #f)
          )
        )
        ; notify the property window, but only if this frame is selected 
        (let
          ((selected (send hierarchy-widget get-selected)))
          (if (equal? selected mred-id)
            (begin
              (send properties-panel set-value 'width w)
              (send properties-panel set-value 'height h)
            )
          )
        )
      )
      
      (define/override (on-subwindow-event w e)
        (let
          ((type (send e get-event-type)))
          (cond 
            ((equal? type 'left-down)
             (set! widget-clicked mred-id)
            )
            ((equal? type 'leave)
             (set! widget-clicked #f)
            )
            ((and (equal? type 'left-up) widget-clicked)
             (if (and preview-callback (procedure? preview-callback))
               (preview-callback widget-clicked)
             )
             (set! widget-clicked #f)
            )
          )
          #f
        )
      )
  
      (super-new)
    )
  )
  
  
  (provide frame-default-values frame-default-preview)
  (define frame-default-values '((type frame)      
                                 (necessary (label))
                                 (label "Untitled Frame")
                                 ; optional
                                 (parent #f)
                                 (width 400)
                                 (height 300)
                                 (x 310)
                                 (y 190)
                                 (style ())
                                 (enabled #t)
                                 (border 0)
                                 (spacing 0)
                                 (alignment (center top))
                                 (min-width 0)
                                 (min-height 0)
                                 (stretchable-width #t)
                                 (stretchable-height #t)))
  (define (frame-default-preview id line parent func)
    (make-object mred-frame%
      id
      func
      (cadr (assoc 'label line))
      parent
      (cadr (assoc 'width line))
      (cadr (assoc 'height line))
      (cadr (assoc 'x line))
      (cadr (assoc 'y line))
      (cadr (assoc 'style line))
      (cadr (assoc 'enabled line))
      (cadr (assoc 'border line))
      (cadr (assoc 'spacing line))
      (cadr (assoc 'alignment line))
      (cadr (assoc 'min-width line))
      (cadr (assoc 'min-height line))
      (cadr (assoc 'stretchable-width line))
      (cadr (assoc 'stretchable-height line)))
  )
  ;;========================================================
  ;; * button
  ;;========================================================
  ; special class for button
  (define mred-button%
    (class button%
      (init-field 
        (mred-id #f)
        (preview-callback #f)
      )
      (define/public (mred-id-set! id)
        (set! mred-id id))
      (define/public (mred-id-get)
        mred-id)
      (define/override (on-subwindow-event w e)
        (let
          ((type (send e get-event-type)))
          (cond 
            ((equal? type 'left-down)
             (set! widget-clicked mred-id)
            )
            ((equal? type 'leave)
             (set! widget-clicked #f)
            )
            ((and (equal? type 'left-up) widget-clicked)
             (if (and preview-callback (procedure? preview-callback))
               (preview-callback widget-clicked)
             )
             (set! widget-clicked #f)
            )
          )
          ; pass the event down
          #f
        )
      )
  
      (super-new)
    )
  )
  
(provide button-default-values button-default-preview)
(define button-default-values (list '(type button)
                                    '(necessary (label parent callback))
                                    '(label "Button")
                                    '(parent #f)
                                     (list 'callback (lambda (button control-event) (void)))
                                     ; optional
                                    '(style ())
                                    '(enabled #t)
                                    '(vert-margin 2)
                                    '(horiz-margin 2)
                                    '(min-width 0)
                                    '(min-height 0)
                                    '(stretchable-width #f)
                                    '(stretchable-height #f)))
  (define (button-default-preview id line parent func)
    (make-object mred-button%
      id
      func
      (cadr (assoc 'label line))
      parent
      (cadr (assoc 'callback line))
      (cadr (assoc 'style line))
      normal-control-font
      (cadr (assoc 'enabled line))
      (cadr (assoc 'vert-margin line))
      (cadr (assoc 'horiz-margin line))
      (cadr (assoc 'min-width line))
      (cadr (assoc 'min-height line))
      (cadr (assoc 'stretchable-width line))
      (cadr (assoc 'stretchable-height line)))
  )
  ;;========================================================
  ;; * radio-box
  ;;========================================================
  ; special class for radio-box
  (define mred-radio-box%
    (class radio-box%
      (init-field 
        (mred-id #f)
        (preview-callback #f)
      )
      (define/public (mred-id-set! id)
        (set! mred-id id))
      (define/public (mred-id-get)
        mred-id)
      (define/override (on-subwindow-event w e)
        (let
          ((type (send e get-event-type)))
          (cond 
            ((equal? type 'left-down)
             (set! widget-clicked mred-id)
            )
            ((equal? type 'leave)
             (set! widget-clicked #f)
            )
            ((and (equal? type 'left-up) widget-clicked)
             (if (and preview-callback (procedure? preview-callback))
               (preview-callback widget-clicked)
             )
             (set! widget-clicked #f)
            )
          )
          ; pass the event down
          #f
        )
      )
      
      (super-new)
    )
  )
  
  (provide radio-box-default-values radio-box-default-preview)
  (define radio-box-default-values (list '(type radio-box)
                                         '(necessary (label choices parent callback))
                                         '(label "Radio Box")
                                         '(choices ("First element" "Second element"))
                                         '(parent #f)
                                         (list 'callback (lambda (radio-box control-event) (void)))
                                         ; optional
                                         '(style (vertical))
                                         '(selection 0)
                                         '(enabled #t)
                                         '(vert-margin 0)
                                         '(horiz-margin 0)
                                         '(min-width 0)
                                         '(min-height 0)
                                         '(stretchable-width #f)
                                         '(stretchable-height #f)))
  (define (radio-box-callback r e)
    (let
      ((selection (send r get-selection)))
      (send hierarchy-widget 
            item-property-set! 
            (send r mred-id-get) 'selection
            selection
            #f
      )
      (send properties-panel set-value 'selection selection)
    )
  )
  (define (radio-box-default-preview id line parent func)
    (make-object mred-radio-box%
      id
      func
      (cadr (assoc 'label line))
      (cadr (assoc 'choices line))
      parent
      radio-box-callback ; I could use this : (cadr (assoc 'callback line))
      (cadr (assoc 'style line))
      (cadr (assoc 'selection line))
      normal-control-font
      (cadr (assoc 'enabled line))
      (cadr (assoc 'vert-margin line))
      (cadr (assoc 'horiz-margin line))
      (cadr (assoc 'min-width line))
      (cadr (assoc 'min-height line))
      (cadr (assoc 'stretchable-width line))
      (cadr (assoc 'stretchable-height line)))
  )
  ;;========================================================
  ;; * check-box
  ;;========================================================
   ; special class for check-box
  (define mred-check-box%
    (class check-box%
      (init-field 
        (mred-id #f)
        (preview-callback #f)
      )
      (define/public (mred-id-set! id)
        (set! mred-id id))
      (define/public (mred-id-get)
        mred-id)
      (define/override (on-subwindow-event w e)
        (let
          ((type (send e get-event-type)))
          (cond 
            ((equal? type 'left-down)
             (set! widget-clicked mred-id)
            )
            ((equal? type 'leave)
             (set! widget-clicked #f)
            )
            ((and (equal? type 'left-up) widget-clicked)
             (if (and preview-callback (procedure? preview-callback))
               (preview-callback widget-clicked)
             )
             (set! widget-clicked #f)
            )
          )
          ; pass the event down
          #f
        )
      )
  
      (super-new)
    )
  )
  
  (provide check-box-default-values check-box-default-preview)
  (define check-box-default-values (list '(type check-box)
                                         '(necessary (label parent callback))
                                         '(label "Check Box")
                                         '(parent #f)
                                         (list 'callback (lambda (check-box control-event) (void)))
                                         ; optional
                                         '(value #f)
                                         '(style ())
                                         '(enabled #t)
                                         '(vert-margin 2)
                                         '(horiz-margin 2)
                                         '(min-width 0)
                                         '(min-height 0)
                                         '(stretchable-width #f)
                                         '(stretchable-height #f)))
  (define (check-box-callback r e)
    (let*
      ((current (send r get-value)))
      (send hierarchy-widget 
            item-property-set! 
            (send r mred-id-get) 'value
            (not current)
            #f
      )
      (if current
        (send properties-panel set-value 'value 0)
        (send properties-panel set-value 'value 1)
      )
    )
  )
  
  (define (check-box-default-preview id line parent func)
    (make-object mred-check-box%
      id
      func
      (cadr (assoc 'label line))
      parent
      check-box-callback ; I could use this : (cadr (assoc 'callback line))
      (cadr (assoc 'style line))
      (cadr (assoc 'value line))
      normal-control-font
      (cadr (assoc 'enabled line))
      (cadr (assoc 'vert-margin line))
      (cadr (assoc 'horiz-margin line))
      (cadr (assoc 'min-width line))
      (cadr (assoc 'min-height line))
      (cadr (assoc 'stretchable-width line))
      (cadr (assoc 'stretchable-height line)))
  )
  ;;========================================================
  ;; * pane
  ;;========================================================
  ; special class for pane
  (define mred-pane%
    (class pane%
      (init-field 
        (mred-id #f)
        (preview-callback #f)
      )
      (define/public (mred-id-set! id)
        (set! mred-id id))
      (define/public (mred-id-get)
        mred-id)
      
      (super-new)
    )
  )
  
  (provide pane-default-values pane-default-preview)
  (define pane-default-values '((type pane)
                                (necessary (parent))
                                (parent #f)
                                ; optional
                                (vert-margin 0)
                                (horiz-margin 0)
                                (border 0)
                                (spacing 0)
                                (alignment (left center))
                                (min-width 0)
                                (min-height 0)
                                (stretchable-width #t)
                                (stretchable-height #t)))
  (define (pane-default-preview id line parent func)
    (make-object mred-pane%
      id
      func
      parent
      (cadr (assoc 'vert-margin line))
      (cadr (assoc 'horiz-margin line))
      (cadr (assoc 'border line))
      (cadr (assoc 'spacing line))
      (cadr (assoc 'alignment line))
      (cadr (assoc 'min-width line))
      (cadr (assoc 'min-height line))
      (cadr (assoc 'stretchable-width line))
      (cadr (assoc 'stretchable-height line)))
  )
  ;;========================================================
  ;; * horizontal-pane
  ;;========================================================
  ; special class for horizontal-pane
  (define mred-horizontal-pane%
    (class horizontal-pane%
      (init-field 
        (mred-id #f)
        (preview-callback #f)
      )
      (define/public (mred-id-set! id)
        (set! mred-id id))
      (define/public (mred-id-get)
        mred-id)
      
      (super-new)
    )
  )
  
  (provide horizontal-pane-default-values horizontal-pane-default-preview)
  (define horizontal-pane-default-values '((type horizontal-pane)
                                           (necessary (parent))
                                           (parent #f)
                                           ; optional
                                           (vert-margin 0)
                                           (horiz-margin 0)
                                           (border 0)
                                           (spacing 0)
                                           (alignment (left center))
                                           (min-width 0)
                                           (min-height 0)
                                           (stretchable-width #t)
                                           (stretchable-height #t)))
  (define (horizontal-pane-default-preview id line parent func)
    (make-object mred-horizontal-pane%
      id
      func
      parent
      (cadr (assoc 'vert-margin line))
      (cadr (assoc 'horiz-margin line))
      (cadr (assoc 'border line))
      (cadr (assoc 'spacing line))
      (cadr (assoc 'alignment line))
      (cadr (assoc 'min-width line))
      (cadr (assoc 'min-height line))
      (cadr (assoc 'stretchable-width line))
      (cadr (assoc 'stretchable-height line)))
  )
  ;;========================================================
  ;; * vertical-pane
  ;;========================================================
  ; special class for vertical-pane
  (define mred-vertical-pane%
    (class vertical-pane%
      (init-field 
        (mred-id #f)
        (preview-callback #f)
      )
      (define/public (mred-id-set! id)
        (set! mred-id id))
      (define/public (mred-id-get)
        mred-id)
  
      (super-new)
    )
  )
  
  (provide vertical-pane-default-values vertical-pane-default-preview)
  (define vertical-pane-default-values '((type vertical-pane)
                                         (necessary (parent))
                                         (parent #f)
                                         ; optional
                                         (vert-margin 0)
                                         (horiz-margin 0)
                                         (border 0)
                                         (spacing 0)
                                         (alignment (left center))
                                         (min-width 0)
                                         (min-height 0)
                                         (stretchable-width #t)
                                         (stretchable-height #t)))
  (define (vertical-pane-default-preview id line parent func)
    (make-object mred-vertical-pane%
      id
      func
      parent
      (cadr (assoc 'vert-margin line))
      (cadr (assoc 'horiz-margin line))
      (cadr (assoc 'border line))
      (cadr (assoc 'spacing line))
      (cadr (assoc 'alignment line))
      (cadr (assoc 'min-width line))
      (cadr (assoc 'min-height line))
      (cadr (assoc 'stretchable-width line))
      (cadr (assoc 'stretchable-height line)))
  )
  ;;========================================================
  ;; * panel
  ;;========================================================
  ; special class for panel
  (define mred-panel%
    (class panel%
      (init-field 
        (mred-id #f)
        (preview-callback #f)
      )
      (define/public (mred-id-set! id)
        (set! mred-id id))
      (define/public (mred-id-get)
        mred-id)
      
      (define/override (on-subwindow-event w e)
        (let
          ((type (send e get-event-type)))
          (cond 
            ((equal? type 'left-down)
             (set! widget-clicked mred-id)
            )
            ((equal? type 'leave)
             (set! widget-clicked #f)
            )
            ((and (equal? type 'left-up) widget-clicked)
             (if (and preview-callback (procedure? preview-callback))
               (preview-callback widget-clicked)
             )
             (set! widget-clicked #f)
            )
          )
          ; pass the event down
          #f
        )
      )
  
      (super-new)
    )
  )
  
  (provide panel-default-values panel-default-preview)
  (define panel-default-values '((type panel)
                                 (necessary (parent))
                                 (parent #f)
                                 ; optional
                                 (style ())
                                 (enabled #t)
                                 (vert-margin 0)
                                 (horiz-margin 0)
                                 (border 0)
                                 (spacing 0)
                                 (alignment (left center))
                                 (min-width 0)
                                 (min-height 0)
                                 (stretchable-width #t)
                                 (stretchable-height #t)))
  (define (panel-default-preview id line parent func)
    (make-object mred-panel%
      id
      func
      parent
      (cadr (assoc 'style line))
      (cadr (assoc 'enabled line))
      (cadr (assoc 'vert-margin line))
      (cadr (assoc 'horiz-margin line))
      (cadr (assoc 'border line))
      (cadr (assoc 'spacing line))
      (cadr (assoc 'alignment line))
      (cadr (assoc 'min-width line))
      (cadr (assoc 'min-height line))
      (cadr (assoc 'stretchable-width line))
      (cadr (assoc 'stretchable-height line)))
  )
  ;;========================================================
  ;; * horizontal-panel
  ;;========================================================
  ; special class for horizontal-panel
  (define mred-horizontal-panel%
    (class horizontal-panel%
      (init-field 
        (mred-id #f)
        (preview-callback #f)
      )
      (define/public (mred-id-set! id)
        (set! mred-id id))
      (define/public (mred-id-get)
        mred-id)
      
      (define/override (on-subwindow-event w e)
        (let
          ((type (send e get-event-type)))
          (cond 
            ((equal? type 'left-down)
             (set! widget-clicked mred-id)
            )
            ((equal? type 'leave)
             (set! widget-clicked #f)
            )
            ((and (equal? type 'left-up) widget-clicked)
             (if (and preview-callback (procedure? preview-callback))
               (preview-callback widget-clicked)
             )
             (set! widget-clicked #f)
            )
          )
          ; pass the event down
          #f
        )
      )
  
      (super-new)
    )
  )
  
  (provide horizontal-panel-default-values horizontal-panel-default-preview)
  (define horizontal-panel-default-values '((type horizontal-panel)
                                            (necessary (parent))
                                            (parent #f)
                                            ; optional
                                            (style ())
                                            (enabled #t)
                                            (vert-margin 0)
                                            (horiz-margin 0)
                                            (border 0)
                                            (spacing 0)
                                            (alignment (left center))
                                            (min-width 0)
                                            (min-height 0)
                                            (stretchable-width #t)
                                            (stretchable-height #t)))
  (define (horizontal-panel-default-preview id line parent func)
    (make-object mred-horizontal-panel%
      id
      func
      parent
      (cadr (assoc 'style line))
      (cadr (assoc 'enabled line))
      (cadr (assoc 'vert-margin line))
      (cadr (assoc 'horiz-margin line))
      (cadr (assoc 'border line))
      (cadr (assoc 'spacing line))
      (cadr (assoc 'alignment line))
      (cadr (assoc 'min-width line))
      (cadr (assoc 'min-height line))
      (cadr (assoc 'stretchable-width line))
      (cadr (assoc 'stretchable-height line)))
  )
  ;;========================================================
  ;; * vertical-panel
  ;;========================================================
  ; special class for pane
  (define mred-vertical-panel%
    (class vertical-panel%
      (init-field 
        (mred-id #f)
        (preview-callback #f)
      )
      (define/public (mred-id-set! id)
        (set! mred-id id))
      (define/public (mred-id-get)
        mred-id)
      
      (define/override (on-subwindow-event w e)
        (let
          ((type (send e get-event-type)))
          (cond 
            ((equal? type 'left-down)
             (set! widget-clicked mred-id)
            )
            ((equal? type 'leave)
             (set! widget-clicked #f)
            )
            ((and (equal? type 'left-up) widget-clicked)
             (if (and preview-callback (procedure? preview-callback))
               (preview-callback widget-clicked)
             )
             (set! widget-clicked #f)
            )
          )
          ; pass the event down
          #f
        )
      )
  
      (super-new)
    )
  )
  
  (provide vertical-panel-default-values vertical-panel-default-preview)
  (define vertical-panel-default-values '((type vertical-panel)
                                          (necessary (parent))
                                          (parent #f)
                                          ; optional
                                          (style ())
                                          (enabled #t)
                                          (vert-margin 0)
                                          (horiz-margin 0)
                                          (border 0)
                                          (spacing 0)
                                          (alignment (left top))
                                          (min-width 0)
                                          (min-height 0)
                                          (stretchable-width #t)
                                          (stretchable-height #t)))
  (define (vertical-panel-default-preview id line parent func)
    (make-object mred-vertical-panel%
      id
      func
      parent
      (cadr (assoc 'style line))
      (cadr (assoc 'enabled line))
      (cadr (assoc 'vert-margin line))
      (cadr (assoc 'horiz-margin line))
      (cadr (assoc 'border line))
      (cadr (assoc 'spacing line))
      (cadr (assoc 'alignment line))
      (cadr (assoc 'min-width line))
      (cadr (assoc 'min-height line))
      (cadr (assoc 'stretchable-width line))
      (cadr (assoc 'stretchable-height line)))
  )
  ;;========================================================
  ;; * tab-panel
  ;;========================================================
  ; special class for tab-pane
  (define mred-tab-panel%
    (class tab-panel%
      (init-field 
        (mred-id #f)
        (preview-callback #f)
      )
      (define/public (mred-id-set! id)
        (set! mred-id id))
      (define/public (mred-id-get)
        mred-id)
      
      (define/override (on-subwindow-event w e)
        (let
          ((type (send e get-event-type)))
          (cond 
            ((equal? type 'left-down)
             (set! widget-clicked mred-id)
            )
            ((equal? type 'leave)
             (set! widget-clicked #f)
            )
            ((and (equal? type 'left-up) widget-clicked)
             (if (and preview-callback (procedure? preview-callback))
               (preview-callback widget-clicked)
             )
             (set! widget-clicked #f)
            )
          )
          ; pass the event down
          #f
        )
      )
  
      (super-new)
    )
  )

  (provide tab-panel-default-values tab-panel-default-preview)
  (define tab-panel-default-values (list '(type tab-panel)
                                         '(necessary (choices parent callback))
                                         '(choices ("First choice" "Second choice"))
                                         '(parent #f)
                                          (list 'callback (lambda (tab-panel control-event) (void)))
                                          ; optional
                                         '(style ())
                                         '(enabled #t)
                                         '(vert-margin 0)
                                         '(horiz-margin 0)
                                         '(border 0)
                                         '(spacing 0)
                                         '(alignment (center top))
                                         '(min-width 0)
                                         '(min-height 0)
                                         '(stretchable-width #t)
                                         '(stretchable-height #t)))
  (define (tab-panel-default-preview id line parent func)
    (make-object mred-tab-panel%
                 id
                 func
                 (cadr (assoc 'choices line))
                 parent
                 (cadr (assoc 'callback line))
                 (cadr (assoc 'style line))
                 normal-control-font
                 (cadr (assoc 'enabled line))
                 (cadr (assoc 'vert-margin line))
                 (cadr (assoc 'horiz-margin line))
                 (cadr (assoc 'border line))
                 (cadr (assoc 'spacing line))
                 (cadr (assoc 'alignment line))
                 (cadr (assoc 'min-width line))
                 (cadr (assoc 'min-height line))
                 (cadr (assoc 'stretchable-width line))
                 (cadr (assoc 'stretchable-height line)))
  )
  ;;========================================================
  ;; * group-box-panel
  ;;========================================================
  ; special class for group-box-panel
  (define mred-group-box-panel%
    (class group-box-panel%
      (init-field 
        (mred-id #f)
        (preview-callback #f)
      )
      (define/public (mred-id-set! id)
        (set! mred-id id))
      (define/public (mred-id-get)
        mred-id)
      
      (define/override (on-subwindow-event w e)
        (let
          ((type (send e get-event-type)))
          (cond 
            ((equal? type 'left-down)
             (set! widget-clicked mred-id)
            )
            ((equal? type 'leave)
             (set! widget-clicked #f)
            )
            ((and (equal? type 'left-up) widget-clicked)
             (if (and preview-callback (procedure? preview-callback))
               (preview-callback widget-clicked)
             )
             (set! widget-clicked #f)
            )
          )
          ; pass the event down
          #f
        )
      )
  
      (super-new)
    )
  )

  (provide group-box-panel-default-values group-box-panel-default-preview)
  (define group-box-panel-default-values '((type group-box-panel)
                                           (necessary (label parent))
                                           (label "Group Box Panel")
                                           (parent #f)
                                           ; optional
                                           (style ())
                                           (enabled #t)
                                           (vert-margin 2)
                                           (horiz-margin 2)
                                           (border 0)
                                           (spacing 0)
                                           (alignment (center top))
                                           (min-width 0)
                                           (min-height 0)
                                           (stretchable-width #t)
                                           (stretchable-height #t)))
  (define (group-box-panel-default-preview id line parent func)
    (make-object mred-group-box-panel%
                 id
                 func
                 (cadr (assoc 'label line))
                 parent
                 (cadr (assoc 'style line))
                 normal-control-font
                 (cadr (assoc 'enabled line))
                 (cadr (assoc 'vert-margin line))
                 (cadr (assoc 'horiz-margin line))
                 (cadr (assoc 'border line))
                 (cadr (assoc 'spacing line))
                 (cadr (assoc 'alignment line))
                 (cadr (assoc 'min-width line))
                 (cadr (assoc 'min-height line))
                 (cadr (assoc 'stretchable-width line))
                 (cadr (assoc 'stretchable-height line)))
  )
  ;;========================================================
  ;; * grow-box-spacer-pane
  ;;========================================================
  ; special class for grow-box-spacer-pane
  (define mred-grow-box-spacer-pane%
    (class grow-box-spacer-pane%
      (init-field 
        (mred-id #f)
        (preview-callback #f)
      )
      (define/public (mred-id-set! id)
        (set! mred-id id))
      (define/public (mred-id-get)
        mred-id)
  
      (super-new)
    )
  )

  (provide grow-box-spacer-pane-default-values grow-box-spacer-pane-default-preview)
  (define grow-box-spacer-pane-default-values '((type grow-box-spacer-pane)
                                                (necessary (parent))
                                                (parent #f)
                                                ; optional
                                                (vert-margin 0)
                                                (horiz-margin 0)
                                                (border 0)
                                                (spacing 0)
                                                (alignment (center center))
                                                (min-width 0)
                                                (min-height 0)
                                                (stretchable-width #t)
                                                (stretchable-height #t)))
  (define (grow-box-spacer-pane-default-preview id line parent func)
    (make-object grow-box-spacer-pane%
                 id
                 func
                 parent
                 (cadr (assoc 'vert-margin line))
                 (cadr (assoc 'horiz-margin line))
                 (cadr (assoc 'border line))
                 (cadr (assoc 'spacing line))
                 (cadr (assoc 'alignment line))
                 (cadr (assoc 'min-width line))
                 (cadr (assoc 'min-height line))
                 (cadr (assoc 'stretchable-width line))
                 (cadr (assoc 'stretchable-height line)))
  )
  ;;========================================================
  ;; * message
  ;;========================================================
  ; special class for message
  (define mred-message%
    (class message%
      (init-field 
        (mred-id #f)
        (preview-callback #f)
      )
      (define/public (mred-id-set! id)
        (set! mred-id id))
      (define/public (mred-id-get)
        mred-id)
      
      (define/override (on-subwindow-event w e)
        (let
          ((type (send e get-event-type)))
          (cond 
            ((equal? type 'left-down)
             (set! widget-clicked mred-id)
            )
            ((equal? type 'leave)
             (set! widget-clicked #f)
            )
            ((and (equal? type 'left-up) widget-clicked)
             (if (and preview-callback (procedure? preview-callback))
               (preview-callback widget-clicked)
             )
             (set! widget-clicked #f)
            )
          )
          ; pass the event down
          #f
        )
      )
  
      (super-new)
    )
  )

  (provide message-default-values message-default-preview)
  (define message-default-values '((type message)
                                   (necessary (label parent))
                                   (label "Message")
                                   (parent #f)
                                   ; optional
                                   (style ())
                                   (enabled #t)
                                   (vert-margin 2)
                                   (horiz-margin 2)
                                   (min-width 0)
                                   (min-height 0)
                                   (stretchable-width #f)
                                   (stretchable-height #f)))
  (define (message-default-preview id line parent func)
    (make-object mred-message%
                 id
                 func
                 (cadr (assoc 'label line))
                 parent
                 (cadr (assoc 'style line))
                 normal-control-font
                 (cadr (assoc 'enabled line))
                 (cadr (assoc 'vert-margin line))
                 (cadr (assoc 'horiz-margin line))
                 (cadr (assoc 'min-width line))
                 (cadr (assoc 'min-height line))
                 (cadr (assoc 'stretchable-width line))
                 (cadr (assoc 'stretchable-height line)))
  )
  ;;========================================================
  ;; * slider
  ;;========================================================
  ; special class for slider
  (define mred-slider%
    (class slider%
      (init-field 
        (mred-id #f)
        (preview-callback #f)
      )
      (define/public (mred-id-set! id)
        (set! mred-id id))
      (define/public (mred-id-get)
        mred-id)
      (define/override (on-subwindow-event w e)
        (let
          ((type (send e get-event-type)))
          (cond 
            ((equal? type 'left-down)
             (set! widget-clicked mred-id)
            )
            ((equal? type 'leave)
             (set! widget-clicked #f)
            )
            ((and (equal? type 'left-up) widget-clicked)
             (if (and preview-callback (procedure? preview-callback))
               (preview-callback widget-clicked)
             )
             (set! widget-clicked #f)
            )
          )
          ; pass the event down
          #f
        )
      )
  
      (super-new)
    )
  )

  (provide slider-default-values slider-default-preview)
  (define slider-default-values (list '(type slider)
                                      '(necessary (label min-value max-value parent callback))
                                      '(label "Slider")
                                      '(min-value 0)
                                      '(max-value 100)
                                      '(parent #f) 
                                       (list 'callback (lambda (slider control-event) (void)))
                                       ; optional
                                      '(init-value 0)
                                      '(style (horizontal)) 
                                      '(enabled #t)
                                      '(vert-margin 2)
                                      '(horiz-margin 2)
                                      '(min-width 0)
                                      '(min-height 0) 
                                      '(stretchable-width #t)
                                      '(stretchable-height #f)))
  (define (slider-callback s e)
    (let
      ((val (send s get-value)))
      (send hierarchy-widget 
            item-property-set! 
            (send s mred-id-get) 'init-value
            val
            #f
      )
      (send properties-panel set-value 'init-value val)
    )
  )
  (define (slider-default-preview id line parent func)
    (make-object mred-slider%
                 id
                 func
                 (cadr (assoc 'label line))
                 (cadr (assoc 'min-value line))
                 (cadr (assoc 'max-value line))
                 parent
                 slider-callback ; I could use this : (cadr (assoc 'callback line))
                 (cadr (assoc 'init-value line))
                 (cadr (assoc 'style line))
                 normal-control-font
                 (cadr (assoc 'enabled line))
                 (cadr (assoc 'vert-margin line))
                 (cadr (assoc 'horiz-margin line))
                 (cadr (assoc 'min-width line))
                 (cadr (assoc 'min-height line))
                 (cadr (assoc 'stretchable-width line))
                 (cadr (assoc 'stretchable-height line)))
  )
  ;;========================================================
  ;; * gauge
  ;;========================================================
  ; special class for gauge
  (define mred-gauge%
    (class gauge%
      (init-field 
        (mred-id #f)
        (preview-callback #f)
      )
      (define/public (mred-id-set! id)
        (set! mred-id id))
      (define/public (mred-id-get)
        mred-id)
      
      (define/override (on-subwindow-event w e)
        (let
          ((type (send e get-event-type)))
          (cond 
            ((equal? type 'left-down)
             (set! widget-clicked mred-id)
            )
            ((equal? type 'leave)
             (set! widget-clicked #f)
            )
            ((and (equal? type 'left-up) widget-clicked)
             (if (and preview-callback (procedure? preview-callback))
               (preview-callback widget-clicked)
             )
             (set! widget-clicked #f)
            )
          )
          ; pass the event down
          #f
        )
      )
  
      (super-new)
    )
  )

  (provide gauge-default-values gauge-default-preview)
  (define gauge-default-values '((type gauge)
                                 (necessary (label range parent))
                                 (label "Gauge")
                                 (range 100)
                                 (parent #f)
                                 ; optional
                                 (style (horizontal))
                                 (enabled #t)
                                 (vert-margin 0)
                                 (horiz-margin 0)
                                 (min-width 0)
                                 (min-height 0)
                                 (stretchable-width #t)
                                 (stretchable-height #f)))
  (define (gauge-default-preview id line parent func)
    (make-object mred-gauge%
                 id
                 func
                 (cadr (assoc 'label line))
                 (cadr (assoc 'range line))
                 parent
                 (cadr (assoc 'style line))
                 normal-control-font
                 (cadr (assoc 'enabled line))
                 (cadr (assoc 'vert-margin line))
                 (cadr (assoc 'horiz-margin line))
                 (cadr (assoc 'min-width line))
                 (cadr (assoc 'min-height line))
                 (cadr (assoc 'stretchable-width line))
                 (cadr (assoc 'stretchable-height line)))
  )
  ;;========================================================
  ;; * list-box
  ;;========================================================
  ; special class for list-box
  (define mred-list-box%
    (class list-box%
      (init-field 
        (mred-id #f)
        (preview-callback #f)
      )
      (define/public (mred-id-set! id)
        (set! mred-id id))
      (define/public (mred-id-get)
        mred-id)
      
      (define/override (on-subwindow-event w e)
        (let
          ((type (send e get-event-type)))
          (cond 
            ((equal? type 'left-down)
             (set! widget-clicked mred-id)
            )
            ((equal? type 'leave)
             (set! widget-clicked #f)
            )
            ((and (equal? type 'left-up) widget-clicked)
             (if (and preview-callback (procedure? preview-callback))
               (preview-callback widget-clicked)
             )
             (set! widget-clicked #f)
            )
          )
          ; pass the event down
          #f
        )
      )
  
      (super-new)
    )
  )

  (provide list-box-default-values list-box-default-preview)
  (define list-box-default-values (list '(type list-box)
                                        '(necessary (label choices parent callback))
                                        '(label "List Box")
                                        '(choices ("First element" "Second element"))
                                        '(parent #f)
                                         (list 'callback (lambda (list-box control-event) (void)))
                                         ; optional
                                        '(style (single))
                                        '(selection #f)
                                        '(enabled #t)
                                        '(vert-margin 0)
                                        '(horiz-margin 0)
                                        '(min-width 0)
                                        '(min-height 0)
                                        '(stretchable-width #t)
                                        '(stretchable-height #t)))
  (define (list-box-default-preview id line parent func)
    (make-object mred-list-box%
                 id
                 func
                 (cadr (assoc 'label line))
                 (cadr (assoc 'choices line))
                 parent
                 (cadr (assoc 'callback line))
                 (cadr (assoc 'style line))
                 (cadr (assoc 'selection line))
                 view-control-font
                 normal-control-font
                 (cadr (assoc 'enabled line))
                 (cadr (assoc 'vert-margin line))
                 (cadr (assoc 'horiz-margin line))
                 (cadr (assoc 'min-width line))
                 (cadr (assoc 'min-height line))
                 (cadr (assoc 'stretchable-width line))
                 (cadr (assoc 'stretchable-height line)))
  )
  ;;========================================================
  ;; * choice
  ;;========================================================
  ; special class for choice
  (define mred-choice%
    (class choice%
      (init-field 
        (mred-id #f)
        (preview-callback #f)
      )
      (define/public (mred-id-set! id)
        (set! mred-id id))
      (define/public (mred-id-get)
        mred-id)
      
      (define/override (on-subwindow-event w e)
        (let
          ((type (send e get-event-type)))
          (cond 
            ((equal? type 'left-down)
             (set! widget-clicked mred-id)
            )
            ((equal? type 'leave)
             (set! widget-clicked #f)
            )
            ((and (equal? type 'left-up) widget-clicked)
             (if (and preview-callback (procedure? preview-callback))
               (preview-callback widget-clicked)
             )
             (set! widget-clicked #f)
            )
          )
          ; pass the event down
          #f
        )
      )
  
      (super-new)
    )
  )

  (provide choice-default-values choice-default-preview)
  (define choice-default-values (list '(type choice)
                                      '(necessary (label choices parent callback))
                                      '(label "Choice")
                                      '(choices ("First element" "Second element"))
                                      '(parent #f)
                                       (list 'callback (lambda (button control-event) (void)))
                                       ; optional
                                      '(style ())
                                      '(selection 0)
                                      '(enabled #t)
                                      '(vert-margin 0)
                                      '(horiz-margin 0)
                                      '(min-width 0)
                                      '(min-height 0)
                                      '(stretchable-width #f)
                                      '(stretchable-height #f)))
  (define (choice-callback r e)
    (let
      ((selection (send r get-selection)))
      (send hierarchy-widget 
            item-property-set! 
            (send r mred-id-get) 'selection
            selection
            #f
      )
      (send properties-panel set-value 'selection selection)
    )
  )
  (define (choice-default-preview id line parent func)
    (make-object mred-choice%
                 id
                 func
                 (cadr (assoc 'label line))
                 (cadr (assoc 'choices line))
                 parent
                 choice-callback ; I could use this : (cadr (assoc 'callback line))
                 (cadr (assoc 'style line))
                 (cadr (assoc 'selection line))
                 normal-control-font
                 (cadr (assoc 'enabled line))
                 (cadr (assoc 'vert-margin line))
                 (cadr (assoc 'horiz-margin line))
                 (cadr (assoc 'min-width line))
                 (cadr (assoc 'min-height line))
                 (cadr (assoc 'stretchable-width line))
                 (cadr (assoc 'stretchable-height line)))
  )
  ;;========================================================
  ;; * text-field
  ;;========================================================
  ; special class for text-field
  (define mred-text-field%
    (class text-field%
      (init-field 
        (mred-id #f)
        (preview-callback #f)
      )
      (define/public (mred-id-set! id)
        (set! mred-id id))
      (define/public (mred-id-get)
        mred-id)
      
      (define/override (on-subwindow-event w e)
        (let
          ((type (send e get-event-type)))
          (cond 
            ((equal? type 'left-down)
             (set! widget-clicked mred-id)
            )
            ((equal? type 'leave)
             (set! widget-clicked #f)
            )
            ((and (equal? type 'left-up) widget-clicked)
             (if (and preview-callback (procedure? preview-callback))
               (preview-callback widget-clicked)
             )
             (set! widget-clicked #f)
            )
          )
          ; pass the event down
          #f
        )
      )
  
      (super-new)
    )
  )

  (provide text-field-default-values text-field-default-preview)
  (define text-field-default-values (list '(type text-field)
                                          '(necessary (label parent callback))
                                          '(label "Text Field")
                                          '(parent #f)
                                           (list 'callback (lambda (text-field control-event) (void)))
                                           ; optional
                                          '(init-value "Text")
                                          '(style (single))
                                          '(enabled #t)
                                          '(vert-margin 2)
                                          '(horiz-margin 2)
                                          '(min-width 0)
                                          '(min-height 0)
                                          '(stretchable-width #t)
                                          '(stretchable-height #f)))
  (define (text-field-default-preview id line parent func)
    (make-object mred-text-field%
                 id
                 func
                 (cadr (assoc 'label line))
                 parent
                 (cadr (assoc 'callback line))
                 (cadr (assoc 'init-value line))
                 (cadr (assoc 'style line))
                 normal-control-font
                 (cadr (assoc 'enabled line))
                 (cadr (assoc 'vert-margin line))
                 (cadr (assoc 'horiz-margin line))
                 (cadr (assoc 'min-width line))
                 (cadr (assoc 'min-height line))
                 (cadr (assoc 'stretchable-width line))
                 (cadr (assoc 'stretchable-height line)))
  )
  
  ;;========================================================
  ;; * canvas
  ;;========================================================
  ; special class for canvas
  (define mred-canvas%
    (class canvas%
      (init-field 
        (mred-id #f)
        (preview-callback #f)
      )
      (define/public (mred-id-set! id)
        (set! mred-id id))
      (define/public (mred-id-get)
        mred-id)
      
      (define/override (on-subwindow-event w e)
        (let
          ((type (send e get-event-type)))
          (cond 
            ((equal? type 'left-down)
             (set! widget-clicked mred-id)
            )
            ((equal? type 'leave)
             (set! widget-clicked #f)
            )
            ((and (equal? type 'left-up) widget-clicked)
             (if (and preview-callback (procedure? preview-callback))
               (preview-callback widget-clicked)
             )
             (set! widget-clicked #f)
            )
          )
          ; pass the event down
          #f
        )
      )
  
      (super-new)
    )
  )

  (provide canvas-default-values canvas-default-preview)
  (define canvas-default-values (list '(type canvas)
                                      '(necessary (parent style))
                                      '(parent #f)
                                      '(style (border))
                                       ; optional
                                       (list 'paint-callback (lambda (canvas control-event) (void)))
                                      '(label "Canvas")
                                      '(gl-config #f)
                                      '(enabled #t)
                                      '(vert-margin 0)
                                      '(horiz-margin 0)
                                      '(min-width 0)
                                      '(min-height 0)
                                      '(stretchable-width #t)
                                      '(stretchable-height #t)))
  (define (canvas-default-preview id line parent func)
    (let*
      ((style    (cadr (assoc 'style line)))
       (hscroll? (if (member 'hscroll style) #t #f))
       (vscroll? (if (member 'vscroll style) #t #f))
       (w (make-object mred-canvas%
                       id
                       func
                       parent
                       (cadr (assoc 'style line))
                       (cadr (assoc 'paint-callback line))
                       (cadr (assoc 'label line))
                       (cadr (assoc 'gl-config line))
                       (cadr (assoc 'enabled line))
                       (cadr (assoc 'vert-margin line))
                       (cadr (assoc 'horiz-margin line))
                       (cadr (assoc 'min-width line))
                       (cadr (assoc 'min-height line))
                       (cadr (assoc 'stretchable-width line))
                       (cadr (assoc 'stretchable-height line)))))
      (send w show-scrollbars hscroll? vscroll?)
      w
    )
  )


) ; end of module

